#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from urllib.request import urlopen
from vrpwrp.helpers import image_helper
from vrpwrp.config import config
from vrpwrp.tools.boundingbox import BoundingBox
from vrpwrp.wrappers.APIWrapper import APIWrapper

__author__ = 'Iván de Paz Centeno'


class FaceDetection(APIWrapper):
    """
    Wrapper for FaceDetection API, from Iván de Paz Centeno API-REST service.
    """
    def __init__(self, API_URL=None):
        """
        API URL for the Face detection algorithm.
        :param API_URL: URL for the face detection algorithm. By default it is going to use the public one.
        :return:
        """
        if API_URL is None:
            API_URL = config.FACE_DETECTION_API

        super().__init__(API_URL)

    def analyze_bytes(self, image_bytes):
        """
        Analyzes an incoming set of raw bytes corresponding to an image file and returns the bounding boxes detected.

        Supported extensions:
            Windows bitmaps - *.bmp, *.dib
            JPEG files - *.jpeg, *.jpg, *.jpe
            JPEG 2000 files - *.jp2
            Portable Network Graphics - *.png
            WebP - *.webp
            Portable image format - *.pbm, *.pgm, *.ppm *.pxm, *.pnm (always supported)
            Sun rasters - *.sr, *.ras
            TIFF files - *.tiff, *.tif
            OpenEXR Image files - *.exr
            Radiance HDR - *.hdr, *.pic

        :param image_bytes: array of bytes of an image.
        :return: list of bounding boxes detected inside the image.
        """

        response = self._request("PUT", data=image_bytes, is_binary=True)['bounding_boxes']

        bounding_boxes = []

        for bbox in response:
            bbox_elements = [int(element) for element in bbox.replace("[","").replace("]", "").split(",")]
            bounding_boxes.append(BoundingBox(*bbox_elements))

        return bounding_boxes

    def analyze_file(self, filename):
        """
        Analyzes a file image and returns the bounding boxes detected.

        Supported extensions:
            Windows bitmaps - *.bmp, *.dib
            JPEG files - *.jpeg, *.jpg, *.jpe
            JPEG 2000 files - *.jp2
            Portable Network Graphics - *.png
            WebP - *.webp
            Portable image format - *.pbm, *.pgm, *.ppm *.pxm, *.pnm (always supported)
            Sun rasters - *.sr, *.ras
            TIFF files - *.tiff, *.tif
            OpenEXR Image files - *.exr
            Radiance HDR - *.hdr, *.pic

        :param filename: URI pointing to the filename to analyze..
        :return: list of bounding boxes detected inside the image.
        """
        image_bytes = image_helper.get_file_binary_content(filename)
        return self.analyze_bytes(image_bytes)

    def analyze_url(self, url):
        """
        Downloads a URL resource and analyzes its content with @analyze_bytes()..

        :param url: url pointing to an image.
        :return: list of bounding boxes detected inside the image.
        """

        image_bytes = urlopen(url).read()
        return self.analyze_bytes(image_bytes)

    def analyze_pil(self, pillow_image):
        """
        Analyzes the content of a Pillow image and returns the detected bounding boxes..

        :param pillow_image: pillow object representing an image..
        :return: list of bounding boxes detected inside the image.
        """

        image_bytes = image_helper.to_byte_array(pillow_image)
        return self.analyze_bytes(image_bytes)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import unittest
from vrpwrp.wrappers.face_detection import FaceDetection

__author__ = 'Iván de Paz Centeno'


class TestFaceDetection(unittest.TestCase):
    """
    Unit tests for the class FaceDetection
    """
    def test_detect_faces_file(self):
        """
        Tests the detection of faces from a file.
        """
        face_detection = FaceDetection()
        bounding_boxes = face_detection.analyze_file("vrpwrp/samples/subject3_3.jpg")
        self.assertEqual(len(bounding_boxes), 1)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import unittest
from vrpwrp.tools.boundingbox import BoundingBox


__author__ = 'Iván de Paz Centeno'


class TestBoundingBox(unittest.TestCase):
    """
    Unitary tests for the BoundingBox class.
    """

    def setUp(self):
        """
        Definition of some common rect values for the tests.
        """
        #                   rect1              rect2                 area_intersection  percentage
        self.rect_sets = [[[80, 60, 250, 170], [200, 130, 200, 170], 13000,             38.24],
                          [[80, 60, 40, 170], [200, 130, 200, 170],  0,                 0.00]]

    def test_expand(self):
        """
        Tests the expansion of bounding box.
        """
        box1 = BoundingBox(3,3,100,100)
        box1.expand() # expand a default of 20%
        self.assertEqual(box1.get_box(), [-7,-7,120,120])   # If you see something weird here,
                                                            # remember that it is the width and height and not coords

    def test_bounding_box_from_string(self):
        """
        Tests the creation a bounding box from a string.
        """
        bbox_string = "22,34,122,432"
        bbox = BoundingBox.from_string(bbox_string)

        self.assertEqual(bbox.get_box(), [22, 34, 122, 432])

        bbox_string = "22, 34,122,432"
        bbox = BoundingBox.from_string(bbox_string)

        self.assertEqual(bbox.get_box(), [22, 34, 122, 432])

        bbox_string = "22, 34, 122, 432"
        bbox = BoundingBox.from_string(bbox_string)

        self.assertEqual(bbox.get_box(), [22, 34, 122, 432])

        bbox_string = "   22, 34, 122, 432   "
        bbox = BoundingBox.from_string(bbox_string)

        self.assertEqual(bbox.get_box(), [22, 34, 122, 432])

        with self.assertRaises(Exception):
            bbox_string = "22,34,122 432"
            bbox = BoundingBox.from_string(bbox_string)

        with self.assertRaises(Exception):
            bbox_string = "22,34  122 432"
            bbox = BoundingBox.from_string(bbox_string)

    def test_fit_in_size(self):
        """
        Tests that bounding box is able to adapt itself to specified bounds.
        """
        image_size = [300, 300]

        # Case1 all out of bounds
        box1 = BoundingBox(-1, -1, 302, 302)
        box1.fit_in_size(image_size)
        self.assertEqual(box1.get_box(), [0, 0, 300, 300])

        # Case2 top left out of bounds
        box1 = BoundingBox(-1,-1,301,301)
        box1.fit_in_size(image_size)
        self.assertEqual(box1.get_box(), [0, 0, 300, 300])

        # Case left out of bounds
        box1 = BoundingBox(-15, 10, 100, 100)
        box1.fit_in_size(image_size)
        self.assertEqual(box1.get_box(), [0, 10, 85, 100])

        # Case top out of bounds
        box1 = BoundingBox(15, -10, 100, 100)
        box1.fit_in_size(image_size)
        self.assertEqual(box1.get_box(), [15, 0, 100, 90])

        # Case width out of bounds
        box1 = BoundingBox(15, 0, 290, 100)
        box1.fit_in_size(image_size)
        self.assertEqual(box1.get_box(), [15, 0, 285, 100])

        # Case height out of bounds
        box1 = BoundingBox(15, 15, 200, 290)
        box1.fit_in_size(image_size)
        self.assertEqual(box1.get_box(), [15, 15, 200, 285])

    def test_get_box(self):
        """
        Tests that bounding box successfully returns the array of the box dimensions.
        """
        box = BoundingBox(15, 16, 17, 18)
        self.assertEqual(box.get_box(), [15, 16, 17, 18])

    def test_get_box_coord(self):
        """
        Tests that bounding box successfully returns the array of the box in coordinates version.
        """
        box = BoundingBox(15, 16, 17, 18)
        self.assertEqual(box.get_box_coord(), [[15, 16], [32, 16], [15, 34], [32, 34]])

    def test_get_components(self):
        """
        Tests that bounding box isolated components are accessible.
        """
        box = BoundingBox(15, 16, 17, 18)
        self.assertEqual(box.get_x(), 15)
        self.assertEqual(box.get_y(), 16)
        self.assertEqual(box.get_width(), 17)
        self.assertEqual(box.get_height(), 18)

    def test_get_numpy_format(self):
        """
        Tests that bounding box numpy format is correct.
        """
        box = BoundingBox(15, 16, 17, 18)
        self.assertEqual(box.get_numpy_format(), [16, 34, 15, 32])

    def test_str(self):
        """
        Tests that bounding box describes itself correctly.
        """
        box = BoundingBox(15, 16, 17, 18)
        self.assertEqual(box.__str__(), "[15, 16, 17, 18]")

    def test_intersection_with_other_boundingbox(self):
        """
        Tests that bounding box intersects itself with other bounding boxes.
        """
        for rect_set in self.rect_sets:
            rect1 = rect_set[0]
            rect2 = rect_set[1]
            expectedArea = rect_set[2]
            expectedPercentage = rect_set[3]

            box1 = BoundingBox(*rect1)
            box2 = BoundingBox(*rect2)

            intersection = box1.intersect_with(box2)
            number_of_common_pixels = intersection.get_area()

            # Which one is the smaller rectangle?
            area_box1 = box1.get_area()
            area_box2 = box2.get_area()

            lesser_area = min(area_box1, area_box2)
            percentage = round((number_of_common_pixels / lesser_area) * 100, 2)

            self.assertEqual(number_of_common_pixels, expectedArea)
            self.assertEqual(percentage, expectedPercentage)

    def test_center(self):
        """
        Tests that bounding box knows its center point.
        """
        box = BoundingBox(10, 10, 20, 20)
        self.assertEqual(box.get_center(), [20, 20])

    def test_area(self):
        """
        Tests that bounding box knows its area.
        """
        box = BoundingBox(10, 10, 20, 20)
        self.assertEqual(box.get_area(), 400)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import io
from PIL import Image
from vrpwrp.tools.boundingbox import BoundingBox

__author__ = 'Iván de Paz Centeno'


HORIZONTAL = 0
VERTICAL = 1

def get_file_binary_content(uri):
    """
    Retrieves the binary content of a file URI.
    :param uri: path to the file to be loaded
    :return: raw bytes list of the file content.
    """
    with open(uri, "rb") as f:
        c = f.read()

    return c

def get_file_image(uri):
    """
    Retrieves the file as a PIL image.
    :param uri: path to the file to be loaded
    :return: PIL image representing the loaded image.
    """
    with Image.open(uri) as im:
        image = im.convert("RGB")
    return image

def get_image(image_bytes):
    """
    Retrieves the PIL image from the given array of bytes.
    :param image_bytes: bytes to convert into a PIL image.
    :return: PIL image.
    """
    with io.BytesIO(image_bytes) as bytes_io, Image.open(bytes_io) as im:
        image = im.convert("RGB")
    return image

def to_byte_array(pil_image):
    """
    converts the PIL image into a bytes array.
    :param pil_image: PIL image to convert to
    :return: Bytes array representing the image.
    """
    with io.BytesIO() as bytes_io:
        pil_image.save(bytes_io, "PNG")
        bytes_io.seek(0)
        result = bytes_io.read()

    return result

def crop_by_bbox(pil_image, bbox):
    """
    Crops the specified PIL image with the given bounding box
    :param pil_image: PIL image to crop
    :param bbox: bounding box object to crop by
    :return: PIL image cropped.
    """
    box = bbox.get_box()
    box[2] += box[0]
    box[3] += box[1]

    crop_result = pil_image.crop((box[0], box[1], box[2], box[3]))

    return crop_result

def get_cropped_faces(pil_image, bounding_boxes):
    """
    Crops the specified PIL image by the specified bounding boxes
    :param pil_image: PIL image to crop by
    :param bounding_boxes: bounding boxes list to crop image.
    :return: A list of cropped images.
    """
    return [crop_by_bbox(pil_image, bounding_box) for bounding_box in bounding_boxes]

def segment_image(pil_image, segment_width, segment_height, iteration_order=HORIZONTAL):
    """
    Generator of segments of a PIL image. This generator breaks the PIL image into a
    subset of PIL images. Like in a Puzzle, it returns one by one on each iteration_order. The iteration_order order can be
     HORIZONTAL (going from TOP-LEFT to TOP-RIGHT) or VERTICAL (going from TOP-LEFT to BOTTOM-LEFT).
    :param pil_image: Image to segment.
    :param segment_width: the width of the fragment.
    :param segment_height: the height of the fragment.
    :param iteration_order: order of the iteration_order.
    :return:
    """
    width, height = pil_image.size

    horizontal_segment_count = width // segment_width + int(width % segment_width > 0)
    vertical_segment_count = height // segment_height + int(height % segment_height > 0)

    if iteration_order==VERTICAL:
        for x in range(horizontal_segment_count):
            for y in range(vertical_segment_count):
                bounding_box = BoundingBox(x * segment_width, y * segment_height,
                                           min(segment_width, width - x * segment_width),
                                           min(segment_height, height - y * segment_height))

                yield crop_by_bbox(pil_image, bounding_box)
    else:
        for y in range(vertical_segment_count):
            for x in range(horizontal_segment_count):
                bounding_box = BoundingBox(x * segment_width, y * segment_height,
                                           min(segment_width, width - x * segment_width),
                                           min(segment_height, height - y * segment_height))

                yield crop_by_bbox(pil_image, bounding_box)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from vrpwrp.config import config
from vrpwrp.helpers import image_helper
from vrpwrp.wrappers.APIWrapper import APIWrapper

__author__ = 'Iván de Paz Centeno'


class PornClassification(APIWrapper):
    def __init__(self, API_URL=None):
        if API_URL is None:
            API_URL = config.PORN_CLASSIFICATION_API

        super().__init__(API_URL)

    def get_score(self, image_bytes):
        try:
            response = self._request("PUT", data=image_bytes, is_binary=True)['Image Result']
            porn_score = float(response['Porn Score'])
        except:
            porn_score = 0.0

        return porn_score

    def get_score_segmented(self, image_bytes, segments_width=1024, segments_height=1024):
        porn_scores = []

        image = image_helper.get_image(image_bytes)

        try:
            for pil_image in image_helper.segment_image(image, segments_width, segments_height):
                response = self._request("PUT", data=image_helper.to_byte_array(pil_image), is_binary=True)['Image Result']
                score = float(response['Porn Score'])
                porn_scores.append(score)
                print(score)

        except:
            pass

        return porn_scores

    def is_porn(self, image_bytes):
        scores = self.get_score_segmented(image_bytes)
        return any([score >= 0.5 for score in scores])


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import unittest
from PIL import Image
from vrpwrp.helpers import image_helper
import io
from vrpwrp.tools.boundingbox import BoundingBox

__author__ = 'Iván de Paz Centeno'


class TestImageHelper(unittest.TestCase):
    """
    Unit tests for the class FaceRecognition
    """

    def setUp(self):
        """
        Sets up the testing subject!
        """
        self.subject = "vrpwrp/samples/subject2_2.jpg" # Hello MR Trump

    def test_get_file_binary_content(self):
        """
        Tests that the image helper can retrieve content from a file.
        """
        content = image_helper.get_file_binary_content(self.subject)

        self.assertGreater(len(content), 0)

        with open(self.subject, "rb") as f:
            original_content = f.read()

        self.assertEqual(content, original_content)

    def test_get_file_image(self):
        """
        Tests that the image helper can retrieve the PIL image from the file.
        """
        image = image_helper.get_file_image(self.subject)

        self.assertEqual(image.size, (800, 450))

    def test_get_image(self):
        """
        Tests that the image helper can retrieve the image from an array of bytes
        """
        with open(self.subject, "rb") as f:
            content = f.read()

        image = image_helper.get_image(content)

        self.assertEqual(image.size, (800, 450))

    def test_to_byte_array(self):
        """
        Tests that the image helper can convert PIL image to a byte array.
        """
        with Image.open(self.subject) as im:
            image = im.convert("RGB")

        byte_array = image_helper.to_byte_array(image)

        self.assertGreater(len(byte_array), 0)

    def test_get_image_and_to_byte_array_are_compatible(self):
        """
        Tests that the helper's got image can be converted to a byte array that can be converted back to an image with same
        format.
        """

        with open(self.subject, "rb") as f:
            content = f.read()

        image = image_helper.get_image(content)

        self.assertEqual(image.size, (800, 450))

        bytes_array = image_helper.to_byte_array(image)

        image = image_helper.get_image(bytes_array)

        self.assertEqual(image.size, (800, 450))

    def test_crop_by_bbox(self):
        """
        Tests that the image helper can crop an image successfully by a bounding box.
        """
        with Image.open(self.subject) as im:
            image = im.convert("RGB")

        cropped = image_helper.crop_by_bbox(image, BoundingBox(0,0,15,15))

        self.assertEqual(cropped.size, (15, 15))

    def test_get_cropped_faces(self):
        """
        Tests that the image helper can crop an image successfully by multiple bounding boxes.
        """
        with Image.open(self.subject) as im:
            image = im.convert("RGB")

        cropped_list = image_helper.get_cropped_faces(image, [BoundingBox(0,0,15,15), BoundingBox(20,20,45,45)])

        self.assertEqual(cropped_list[0].size, (15, 15))
        self.assertEqual(cropped_list[1].size, (45, 45))

    def test_segment_image(self):
        """
        Tests that the image helper can segment an image successfully.
        """
        with Image.open(self.subject) as im:
            image = im.convert("RGB")

        segment_generator = image_helper.segment_image(image, 200, 200)

        sizes = [
            ((200, 200), next(segment_generator)),
            ((200, 200), next(segment_generator)),
            ((200, 200), next(segment_generator)),
            ((200, 200), next(segment_generator)),
            ((200, 200), next(segment_generator)),
            ((200, 200), next(segment_generator)),
            ((200, 200), next(segment_generator)),
            ((200, 200), next(segment_generator)),
            ((200, 50), next(segment_generator)),
            ((200, 50), next(segment_generator)),
            ((200, 50), next(segment_generator)),
            ((200, 50), next(segment_generator))
        ]

        for size, segment in sizes:
            self.assertEqual(size, segment.size)


if __name__ == '__main__':
    unittest.main()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = 'Iván de Paz Centeno'

FACE_DETECTION_API = "http://visnode.sockhost.net:9909/detection-requests/faces/stream"
FACE_RECOGNITION_API = "http://visnode.sockhost.net:9909/recognition-requests/face/embedding/stream"
PORN_CLASSIFICATION_API = ""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import json
from urllib.request import urlopen
from vrpwrp.helpers import image_helper
from vrpwrp.config import config
from vrpwrp.tools.embedding import Embedding, NUMPY_LOADED
from vrpwrp.wrappers.APIWrapper import APIWrapper

__author__ = 'Iván de Paz Centeno'

NUMPY_AVAILABLE = NUMPY_LOADED


class FaceRecognition(APIWrapper):
    """
    Wrapper for FaceRecognition API, from Iván de Paz Centeno API-REST service.
    """
    def __init__(self, API_URL=None):
        """
        API URL for the Face recognition algorithm.
        :param API_URL: URL for the face recognition algorithm. By default it is going to use the public one.
        :return:
        """
        if API_URL is None:
            API_URL = config.FACE_RECOGNITION_API

        super().__init__(API_URL)

    def get_embeddings_from_bytes(self, image_bytes):
        """
        Retrieves the face embedding for the given raw array of bytes of an image file.

        Supported extensions:
            Windows bitmaps - *.bmp, *.dib
            JPEG files - *.jpeg, *.jpg, *.jpe
            JPEG 2000 files - *.jp2
            Portable Network Graphics - *.png
            WebP - *.webp
            Portable image format - *.pbm, *.pgm, *.ppm *.pxm, *.pnm (always supported)
            Sun rasters - *.sr, *.ras
            TIFF files - *.tiff, *.tif
            OpenEXR Image files - *.exr
            Radiance HDR - *.hdr, *.pic

        :param image_bytes: array of bytes of the image to process.
        :return: embedding string representing the image of the face.
        """
        response = self._request("GET", data=image_bytes, is_binary=True)['embedding_data']
        embedding = Embedding(response['embedding'], self)
        return embedding

    def get_embeddings_from_file(self, filename):
        """
        Analyzes a file image and returns the embeddings of the face.

        Supported extensions:
            Windows bitmaps - *.bmp, *.dib
            JPEG files - *.jpeg, *.jpg, *.jpe
            JPEG 2000 files - *.jp2
            Portable Network Graphics - *.png
            WebP - *.webp
            Portable image format - *.pbm, *.pgm, *.ppm *.pxm, *.pnm (always supported)
            Sun rasters - *.sr, *.ras
            TIFF files - *.tiff, *.tif
            OpenEXR Image files - *.exr
            Radiance HDR - *.hdr, *.pic

        :param filename: URI pointing to the filename to analyze..
        :return: embedding string representing the image of the face.
        """
        image_bytes = image_helper.get_file_binary_content(filename)
        return self.get_embeddings_from_bytes(image_bytes)

    def get_embeddings_from_url(self, url):
        """
        Retrieves the face embedding for the image hosted in the given url.

        :param url: url pointing to an image.
        :return: embedding string representing the image of the face.
        """
        image_bytes = urlopen(url).read()
        return self.get_embeddings_from_bytes(image_bytes)

    def get_embeddings_from_pil(self, pillow_image):
        """
        Retrieves the face embedding for the given pillow image.

        :param pillow_image: pillow image to process..
        :return: embedding string representing the image of the face.
        """
        image_bytes = image_helper.to_byte_array(pillow_image)
        return self.get_embeddings_from_bytes(image_bytes)

    def get_embeddings_distance(self, embedding1, embedding2):
        """
        Computes the distance between two embeddings. The distance is a float number.
        The closest the distance (nearest to 0), the most likely to belong to the same face.
        Generally it can be considered that a distance over 1.0 might indicate that the faces are from different
        persons.

        :param embedding1: embedding string for face 1.
        :param embedding2: embedding string for face 2.
        :return: the distance between both embeddings, in float format.
        """
        return self.get_embeddings_distances(embedding1, [embedding2])[0]

    def get_embeddings_distances(self, embedding_who, embeddings_list):
        """
        Computes the distances between an embedding and a list of embeddings. This method is optimal for processing a
        comparison against multiple embeddings, rather than going in a loop one by one.
        :param embedding_who: embedding that wants to be compared.
        :param embeddings_list: the list of embeddings to compare to.
        :return: an array of the distances between the embedding_who and each of the embeddings in the embeddings_list.
        """
        global NUMPY_AVAILABLE

        if NUMPY_AVAILABLE:
            result = [embedding_who - emb for emb in embeddings_list]
        else:
            data={'who': str(embedding_who), 'subjects': [str(embedding) for embedding in embeddings_list]}
            values = self._request("PUT", data=json.dumps(data))['distances']
            result = [float(val) for val in values]

        return result




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import requests

__author__ = 'Iván de Paz Centeno'


class APIWrapper(object):

    def __init__(self, API_URL):
        self.API_URL = API_URL

    def _request(self, method, params=None, data=None, is_binary=False):
        methods = {
            'GET': requests.get,
            'PUT': requests.put
        }

        if not is_binary and data is not None:
            data = data
            headers = {'content-type': 'application/json'}
        else:
            headers = {}

        response = methods[method](self.API_URL, params=params, data=data, headers=headers)
        return response.json()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sys
from setuptools import setup, setuptools

__author__ = 'Iván de Paz Centeno'

def readme():
    with open('README.rst') as f:
        return f.read()

if sys.version_info < (3, 4, 1):
    sys.exit('Python < 3.4.1 is not supported!')

setup(name='vrpwrp',
      version='0.0.5',
      description='Vision-algorithms Requests Processing Wrappers for deep-learning Computer Vision algorithms on the cloud.',
      long_description=readme(),
      url='http://github.com/ipazc/vrpwrp',
      author='Iván de Paz Centeno',
      author_email='ipazc@unileon.es',
      license='MIT',
      packages=setuptools.find_packages(),
      install_requires=[
          'requests',
          'pillow'
      ],
      test_suite='nose.collector',
      tests_require=['nose'],
      include_package_data=True,
      keywords="vrpwrp face_detection face_recognition face deep-learning computer vision face detection face recognition api rest wrapper",
      zip_safe=False)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import json

__author__ = 'Iván de Paz Centeno'


class BoundingBox(object):
    """
    Represents a virtual bounding box and offers some basics operations on it, like
    calculating the area, the intersection with other bounding box or the center.
    """

    def __init__(self, x, y, width, height):
        """
        Builds a bounding box object wrapping the coordinates of the top-left corner of the rect and its width and
        height.

        :param x: top-left x coordinate
        :param y: top-left y coordinate
        :param width: size of the width.
        :param height: size of the height.
        """
        self.x = x
        self.y = y
        self.width = width
        self.height = height

    def get_box(self):
        """
        :return:
        Returns the top-left position of the box and its width and height, as a vector of 4 positions.
        """
        return [self.x, self.y, self.width, self.height]

    def get_box_coord(self):
        """
        Computes the coordinates of the four corners of the rect.
        The coordinates of each corner are wrapped inside a vector.


        :return: the coordinates of each corner of the rect as a vector of 4 positions.
        """
        return [[self.x, self.y], [self.x+self.width, self.y],
                [self.x, self.y+self.height], [self.x+self.width, self.y+self.height]]

    def get_x(self):
        """
        Getter for x coord
        :return:
        """
        return self.x

    def get_y(self):
        """
        Getter for y coord
        :return:
        """
        return self.y

    def get_width(self):
        """
        Getter for width
        :return:
        """
        return self.width

    def get_height(self):
        """
        Getter for height
        :return:
        """
        return self.height

    def get_numpy_format(self):
        """
        Computes a numpy format.
        :return: array of [y, y+height, x, x+width]
        """
        return [self.y, self.y+self.height, self.x, self.x + self.width]

    def intersect_with(self, other_bounding_box):
        """
        Builds the intersection box from this box with the specified bounding box.

        :param other_bounding_box: the bounding box to intersect with.
        :return: the bounding box that corresponds to the intersection box. If the specified box does not intersect
        with the current box, the result will be a bounding box with an area of 0.
        """
        bbox1_coords = self.get_box_coord()
        bbox2_coords = other_bounding_box.get_box_coord()

        bbox1_first_coord = bbox1_coords[0]
        bbox2_first_coord = bbox2_coords[0]

        bbox1_last_coord = bbox1_coords[3]
        bbox2_last_coord = bbox2_coords[3]

        intersection_first_coord = [0, 0]
        intersection_last_coord = [0, 0]

        if bbox1_first_coord[0] >= bbox2_first_coord[0]:
            intersection_first_coord[0] = bbox1_first_coord[0]
        else:
            intersection_first_coord[0] = bbox2_first_coord[0]

        if bbox1_first_coord[1] >= bbox2_first_coord[1]:
            intersection_first_coord[1] = bbox1_first_coord[1]
        else:
            intersection_first_coord[1] = bbox2_first_coord[1]

        if bbox1_last_coord[0] <= bbox2_last_coord[0]:
            intersection_last_coord[0] = bbox1_last_coord[0]
        else:
            intersection_last_coord[0] = bbox2_last_coord[0]

        if bbox1_last_coord[1] <= bbox2_last_coord[1]:
            intersection_last_coord[1] = bbox1_last_coord[1]
        else:
            intersection_last_coord[1] = bbox2_last_coord[1]

        # condition: intersectionFirstCoord never can be greater than intersectionLastCoord
        if intersection_first_coord[0] >= intersection_last_coord[0] \
                or intersection_first_coord[1] >= intersection_last_coord[1]:
            intersection_bounding_box = BoundingBox(0, 0, 0, 0)
        else:
            intersection_bounding_box = BoundingBox(intersection_first_coord[0], intersection_first_coord[1],
                                                    intersection_last_coord[0] - intersection_first_coord[0],
                                                    intersection_last_coord[1] - intersection_first_coord[1])

        return intersection_bounding_box

    def get_area(self):
        """
            Computes the area of the current box.

        :return: the area of the current box.
        """
        return self.width*self.height

    def get_center(self):
        """
        Computes the center of the bounding box
        :return: array (point 2D) of the center.
        """
        return [int(self.x + self.width/2), int(self.y + self.height/2)]

    def expand(self, proportion=0.2):
        """
        Expands the current box area in a given proportion.
        The size expanded is applied to all the limits of the box

        WARNING: The box may end up out of bounds. Call fit_in_size with the boundings to adapt it.
        :param proportion:
        """
        horizontally = int((self.width * proportion) / 2)
        vertically = int((self.height * proportion) / 2)
        self.x -= horizontally
        self.y -= vertically
        self.width += horizontally * 2
        self.height += vertically * 2

    def fit_in_size(self, size_limit):
        """
        Adapts the size of the box in order to avoid exceeding the bounds specified in size_limit

        :param size_limit: bound limits. For example [300, 400] for a maximum width of 300 and 400 for maximumg height.
        """
        if self.x < 0:
            self.width += self.x   # self.x is negative
            self.x = 0

        elif self.x >= size_limit[0]: # Invalid box
            self.x = 0
            self.width = 0

        if self.x + self.width >= size_limit[0]:
            self.width = size_limit[0] - self.x

        if self.y < 0:
            self.height += self.y   # self.y is negative
            self.y = 0

        elif self.y >= size_limit[1]: # Invalid box
            self.y = 0
            self.height = 0

        if self.y + self.height >= size_limit[1]:
            self.height = size_limit[1] - self.y

    def to_dict(self):
        """
        :return: JSON-Compatible dictionary representation of the age.
        """
        return [self.x, self.y, self.width, self.height]

    def __str__(self):
        """
        :return: string representation of the age.
        """
        return json.dumps(self.to_dict())

    @classmethod
    def from_string(cls, bounding_box_string):
        """
        Creates an instance of the bounding box by giving a string.
        :param bounding_box_string: string of the format "x,y,width,height"
        :return: instance of boundingbox matching the string.
        """
        bbox_elements = bounding_box_string.split(",")

        if len(bbox_elements) != 4:
            raise Exception("bounding box string doesn't contain a bounding box.")

        x = int(bbox_elements[0].strip())
        y = int(bbox_elements[1].strip())
        width = int(bbox_elements[2].strip())
        height = int(bbox_elements[3].strip())

        result = cls(x, y, width, height)

        return result


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Numpy can be used to speed up comparison of embeddings.
try:
    import numpy as np
    NUMPY_LOADED = True
except Exception as ex:
    NUMPY_LOADED = False

__author__ = 'Iván de Paz Centeno'


class Embedding(object):
    """
    Represents the embeddings for a face.
    It is a set of float numbers that identifies a face.
    """

    def __init__(self, np_embedding, face_recognition=None):
        """
        Initializer for the embedding object.

        :param np_embedding: numpy array, list or string representing the numpy array of the embeddings.
        :param face_recognition: face recognition object to use when no numpy library is available. If not specified, it
        is going to use the default one.
        :return:
        """
        emb_type = type(np_embedding)

        if emb_type is str:
            if NUMPY_LOADED:
                self.np_embedding = np.fromstring(np_embedding.replace("[","").replace("]", ""), sep=" ")
            else:
                self.np_embedding = np_embedding

        elif NUMPY_LOADED and emb_type is np.ndarray:
            self.np_embedding = np_embedding

        elif emb_type is list:
            if NUMPY_LOADED:
                self.np_embedding = np.asarray(np_embedding)
            else:
                self.np_embedding = str(np_embedding).replace(",", " ")
        else:
            raise Exception("The NP-Embeddings must be a numpy array, a list of floats or a string representation.")

        if face_recognition is None:
            from vrpwrp.wrappers.face_recognition import FaceRecognition
            face_recognition = FaceRecognition()

        self.face_recognition = face_recognition

    @classmethod
    def from_dict(cls, dict_rep, face_recognition=None):
        """
        Builds the embedding from the dictionary representation.
        :param dict_rep:
        :return:
        """
        return cls(dict_rep['embedding'], face_recognition)

    def __sub__(self, other):
        """
        Subtracts other embedding to this embedding.
        :param other: Embedding object
        :return: Embedding object with the subtraction result.
        """
        if NUMPY_LOADED:
            result = float(np.sqrt(np.sum(np.square(np.subtract(self.np_embedding, other.np_embedding)))))
        else:
            result = self.face_recognition.get_embeddings_distance(self, other)

        return float(result)

    def get_embedding_np(self):
        """
        returns the internal NP_Embedding value. It might be a numpy ND-Array or a string depending on if Numpy is
        available or not.
        :return:
        """
        return self.np_embedding

    def to_dict(self):
        """
        Crafts a dictionary representation of this embedding.
        :return: dict representation of the embedding.
        """
        return {"embedding": str(self.np_embedding)}

    def __str__(self):
        return str(self.get_embedding_np())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from urllib.request import urlopen
from vrpwrp.helpers import image_helper
from vrpwrp.wrappers.face_detection import FaceDetection
from vrpwrp.wrappers.face_recognition import FaceRecognition

__author__ = 'Iván de Paz Centeno'


class FaceBundle(object):

    def __init__(self, face_detection=None, face_recognition=None):
        if face_detection is None:
            face_detection = FaceDetection()

        if face_recognition is None:
            face_recognition = FaceRecognition()

        self.face_detection = face_detection
        self.face_recognition = face_recognition

    def process_file(self, uri):
        image_bytes = image_helper.get_file_binary_content(uri)
        bounding_boxes = self.face_detection.analyze_bytes(image_bytes)
        image = image_helper.get_image(image_bytes)
        cropped_images = [image_helper.crop_by_bbox(image, bb) for bb in bounding_boxes]
        embeddings = [self.face_recognition.get_embedding_from_pil(cropped).get_embedding_np() for cropped in cropped_images]

        return embeddings

    def process_url(self, url):
        image_bytes = urlopen(url).read()
        bounding_boxes = self.face_detection.analyze_bytes(image_bytes)
        image = image_helper.get_image(image_bytes)
        cropped_images = [image_helper.crop_by_bbox(image, bb) for bb in bounding_boxes]
        embeddings = [self.face_recognition.get_embedding_from_pil(cropped).get_embedding_np() for cropped in cropped_images]

        return embeddings